/*
Copyright 2006, Freie Universitaet Berlin. All rights reserved.

These sources were developed at the Freie Universitt Berlin, Computer
Systems and Telematics group.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

- Redistributions of source code must retain the above copyright
notice, this list of conditions and the following disclaimer.

- Redistributions in binary form must reproduce the above copyright
notice, this list of conditions and the following disclaimer in the
documentation and/or other materials provided with the distribution.
 
- Neither the name of Freie Universitaet Berlin (FUB) nor the names of its
contributors may be used to endorse or promote products derived from
this software without specific prior written permission.

This software is provided by FUB and the contributors on an "as is"
basis, without any representations or warranties of any kind, express
or implied including, but not limited to, representations or
warranties of non-infringement, merchantability or fitness for a
particular purpose. In no event shall FUB or contributors be liable
for any direct, indirect, incidental, special, exemplary, or
consequential damages (including, but not limited to, procurement of
substitute goods or services; loss of use, data, or profits; or
business interruption) however caused and on any theory of liability,
whether in contract, strict liability, or tort (including negligence
or otherwise) arising in any way out of the use of this software, even
if advised of the possibility of such damage.

This implementation was developed by the CST group at the FUB.

For documentation and questions please use the web site
http://scatterweb.mi.fu-berlin.de and the mailinglist
scatterweb@lists.spline.inf.fu-berlin.de (subscription via the Website).
Berlin, 2006
*/

#include <ScatterWeb.System.h>
#include <ScatterWeb.Sd.h>

#define LED_ON			P5OUT &=~0x80
#define LED_OFF			P5OUT |= 0x80
#define LED_TOGGLE		P5OUT ^= 0x80

char					buffer[512];
sd_cache_t				cache;

uint8_t	global_ret;
uint16_t	global_r2;

bool Process_radioHandler(struct netpacket_handler_args* args)
{
	/*
		Notes: * The received packet data becomes invalid when this handler is completed.
	*/

	// Pass packet through command interface
	if ( Cmd_radioHandler( args ) ) {
		// packet handled by messaging module
		return true;
	}

	if ( args->cpaResult == ADRCOMP_OTHER ) {
		// intercept packets for other nodes here
	} else if ( args->cpaResult == ADRCOMP_BROADCAST ) {
		// process broadcast packets here
	} else {
		// process packets for this node here
		switch ( args->netheader->type ) {

			default:
				return false;		// packet not handled (unknown)
		}
		return true;				// packet handled
	}
	return false;					// packet not handled (other destination)
}

void Process_sensorHandler(void* sdata) {

}

// memory info
COMMAND(sdm, 0, cmdargs) {
	uint32_t		size = 0;
	uint16_t		mb, kb;

	// read size from csd register
	size = Sd_get_size();

	// calculate card size in megabytes from bytes
	size >>= 10;
	kb = size % 1024;
	size >>= 10;
	mb = size;

	sprintf(cmdargs->response_args, "Supp. blocklength: 2^%u-2^%u size: %u.%u MiB", sd_state.MinBlockLen_bit, sd_state.MaxBlockLen_bit, mb, kb);
}

// manufacturer info
COMMAND(sdi, 0, cmdargs) {
	struct sd_cid	cid;
	
	if( Sd_read_cid(&cid) ) {
		snprintf(cmdargs->response_args, cmdargs->response_bufsize,
			"CID mid:%#.2x, oid:%.2s, pnm:%.5s, psn:%lu, rev:%#.2x, mdt:%#.4x",
			cid.mid, cid.oid, cid.pnm, cid.psn, cid.revision, cid.mdt);
	}
}

// set blocklength
COMMAND(sdl, 0, cmdargs) {
	uint16_t bl = String_parseUint16(cmdargs->args, NULL);

	if( bl > sd_state.MaxBlockLen_bit || bl < sd_state.MinBlockLen_bit ) {
		snprintf(cmdargs->response_args, cmdargs->response_bufsize,
		"ERROR: Supp. blocklength 2^%u-2^%u", sd_state.MinBlockLen_bit, sd_state.MaxBlockLen_bit);
	} else {
		bl = Sd_set_blocklength(bl);
		snprintf(cmdargs->response_args, cmdargs->response_bufsize, "blocklength: 2^%u", sd_state.BlockLen_bit);
	}
}

COMMAND(sd1, CMDFLAG_SERIAL, cmdargs) {
	uint32_t address = strtoul(cmdargs->args, NULL, 0);
	uint16_t ret;

	char x[20];

	ret = Sd_read(x, address, 20);
	if( ret ) {
		uint8_t	cs[2];
		uint16_t	i;

		printf("%u Bytes follow:\r\n", 20);
		for(i = 0; i < 20; ) {
			String_fromByteToHex(x[i], cs);
			printf("%.2s", cs);
			i++;
			if( (i % 2) == 0 ) printf(" ");
			if( (i % 16) == 0 ) printf("\r\n");
		}
	} else {
		printf("ERROR?");
	}
}


// read block
COMMAND(sdr, CMDFLAG_SERIAL, cmdargs) {
	uint32_t address = strtoul(cmdargs->args, NULL, 0);
	uint16_t ret;

	ret = Sd_read_block(buffer, address);
	if( ret ) {
		uint8_t	cs[2];
		uint16_t	i;

		printf("%u Bytes follow:\r\n", sd_state.BlockLen);
		for(i = 0; i < sd_state.BlockLen; ) {
			String_fromByteToHex(buffer[i], cs);
			printf("%.2s", cs);
			i++;
			if( (i % 2) == 0 ) printf(" ");
			if( (i % 16) == 0 ) printf("\r\n");
		}
	} else {
		printf("ERROR: Address not block aligned?");
	}
}

// write n blocks
COMMAND(sdw, CMDFLAG_SERIAL, cmdargs) {
	uint8_t* c = (uint8_t*)cmdargs->args;

	uint16_t ba = String_parseUint16(c, &c);
	while(*c == ' ') c++;
	uint16_t count = String_parseUint16(c, &c);

	enum sd_write_ret ret;
	uint16_t i;

	uint8_t x = 0;		// byte to fill blocks with
	uint8_t retry = 0;	// retry count

	uint32_t address = ba << sd_state.BlockLen_bit;
	printf("Writing %u blocks of %u bytes\r\n", count, sd_state.BlockLen);
	for( i = 1; i <= count; i++, address += sd_state.BlockLen, x++ ) {
		ret = Sd_set_block(address, &x);
		printf("\t* block %u (retry %u) set to %#.2X at %lu\r\n", i, retry, x, address);
		if( ret == SD_WRITE_SUCCESS ) {
			retry = 0;
		} else if( ret == SD_WRITE_STORE_ERR && retry < 1 ) {
			printf("\t\tstoring of written data failed\r\n");
			address -= sd_state.BlockLen;
			x--;
			i--;
			retry++;
		} else {
			printf("\t\terror writing block (%u)\r\n", i);
			printf("SD store failed (ret(%#.2X),r2(%#.4X)", global_ret, global_r2);
			return;
		}
		
		// reset watchdog
		System_startWatchdog();
	}
}

COMMAND(sdb, 0, cmdargs) {
	const int bl = 9;							// blocklength
	int mb = String_parseUint16(cmdargs->args, NULL);		// num MB to read
	int n = mb << (20 - bl);					// num blocks to read
	int i;
	double bpsd;
	uint32_t bps;

	time_t startt, endt;
	uint32_t duration;
	uint32_t address = 0;

	if( mb < 1 ) {
		sprintf(cmdargs->response_args, "enter size to read in MB as param");
		return;
	}
	printf("reading %u MiB (wait about 10 s/MB )", mb);

	uart_lock_wait(UART_MODE_SPI);
	Sd_set_blocklength(bl);

	System_stopWatchdog();
	_DINT();
	startt = CURRENT_TIME;
	for( i = 0; i < n; i++ ) {
		Sd_read_block(buffer, address);
		address += (1 << bl);
	}
	endt = CURRENT_TIME;
	_EINT();
	System_startWatchdog();

	duration = (endt.secs - startt.secs) * 1000;
	duration += endt.millis;
	duration -= startt.millis;

	bpsd = duration;
	bpsd /= 1000;
	bpsd = ((mb<<10) / bpsd) * 1024;
	bps = (uint32_t)bpsd;

	uart_unlock(UART_MODE_SPI);
	LOG_LOW("DCOCTL: %.2x, BCSCTL1: %.2x BCSCTL2: %.2x\r\n", DCOCTL, BCSCTL1, BCSCTL2);
	
	sprintf(cmdargs->response_args, " took approx. %lu ms (%lu kbps)", duration, bps);
}

void Sd_Initialize() {
	bool ret = 0;

	// initialize card before anything else
	ret = Sd_init_card(&cache);
	
	// result of initialization and card check
	printf("SD Initialized result: %u (> 0 is error)\r\n", ret);
	printf("Type\t'sdm'\t\tfor card size\r\n");
	printf("Type\t'sdi'\t\tfor card information\r\n");
	printf("Type\t'sdl i'\t\tto set blocklength to 2^i\r\n");
	printf("Type\t'sdr a'\t\tto read a block from block aligned address a\r\n");
	printf("Type\t'sdw ba n'\tto write n block from block address ba\r\n");
	printf("\t\t\tWARNING: Overwrites any existing data\r\n");
	printf("Type\t'sdb n'\t\t to read n megabytes and report duration\r\n");
}

void Process_init() {
	System_registerCallback(C_RADIO, (fp_vp)Process_radioHandler);
	System_registerCallback(C_SENSOR, (fp_vp)Process_sensorHandler);

	// enable verbose logging
	Debuglog_level = VERBOSE;

	Sd_Initialize();
}

// Software header.
APP(
	APPH_STATIC_APP,	// Application Type
	__APP_NAME,			// Application Name (8 chars)
	1, 0				// Aplication Version (2 bytes)
);
