/*
Copyright 2007, Freie Universitaet Berlin. All rights reserved.

These sources were developed at the Freie Universitt Berlin, Computer
Systems and Telematics group.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

- Redistributions of source code must retain the above copyright
notice, this list of conditions and the following disclaimer.

- Redistributions in binary form must reproduce the above copyright
notice, this list of conditions and the following disclaimer in the
documentation and/or other materials provided with the distribution.

- Neither the name of Freie Universitaet Berlin (FUB) nor the names of its
contributors may be used to endorse or promote products derived from
this software without specific prior written permission.

This software is provided by FUB and the contributors on an "as is"
basis, without any representations or warranties of any kind, express
or implied including, but not limited to, representations or
warranties of non-infringement, merchantability or fitness for a
particular purpose. In no event shall FUB or contributors be liable
for any direct, indirect, incidental, special, exemplary, or
consequential damages (including, but not limited to, procurement of
substitute goods or services; loss of use, data, or profits; or
business interruption) however caused and on any theory of liability,
whether in contract, strict liability, or tort (including negligence
or otherwise) arising in any way out of the use of this software, even
if advised of the possibility of such damage.

This implementation was developed by the CST group at the FUB.

For documentation and questions please use the web site
http://scatterweb.mi.fu-berlin.de and the mailinglist
scatterweb@lists.spline.inf.fu-berlin.de (subscription via the Website).
Berlin, 2007
*/

/**
	Reads the SHT11 sensor in intervals and sends result to sink node.
	Sink node prints results.
	Can be flashed to both sensor nodes and sink. Configuration below.
*/

#include <ScatterWeb.System.h>
#include <ScatterWeb.CC1020.h>
#include <ScatterWeb.SHT11.h>
#include <string.h>

struct sht11_data {
	float temperature;
	float humidity;
} data;

extern void ftoa(uint8_t* buf, double f, uint8_t pad);
void Send_Sht11();
void Print_Sht11(netaddr_t* source);

#define sht11_data_packet_type	(USERDEFINED_PACKET + 1)
// --------------- SINK NODE CONFIGURATION --------------
#define GATEWAY_NODE	42
// --------------- READ-OUT INTERVAL --------------------
#define SHT11_INTERVAL	5000

bool Process_radioHandler(struct netpacket_handler_args* args)
{
	/*
		Notes: * The received packet data becomes invalid when this handler is completed.
	*/

	// Pass packet through command interface
	if ( Cmd_radioHandler( args ) ) {
		// packet handled by messaging module
		return true;
	}

	if ( args->cpaResult == ADRCOMP_OTHER ) {
		// intercept packets for other nodes here
	} else if ( args->cpaResult == ADRCOMP_BROADCAST ) {
		// process broadcast packets here
	} else {
		// process packets for this node here
		switch ( args->netheader->type ) {
			case sht11_data_packet_type:
				if( args->payload_size == sizeof(struct sht11_data) ) {
					memcpy(&data, args->payload, sizeof(struct sht11_data));
					Print_Sht11(&args->netheader->from);
				}
				break;
				
			case PING_REQUEST_PACKET:
				// send back to receiver's address
				Net_sendPong(args->netheader->from);
				break;

			case PING_REPLY_PACKET:
				LOG_LOW( "PONG from %u", args->netheader->from );
				break;

			default:
				return false;		// packet not handled (unknown)
		}
		return true;				// packet handled
	}
	return false;					// packet not handled (other destination)
}

void Process_sensorHandler(void* sdata)
{

}

void Send_Sht11() {
	netpacket_send_args_t npsargs;

	data.temperature = SHT11_temperature(SHT11_celsius);
	data.humidity = SHT11_humidity();
	
	if( Configuration.id != GATEWAY_NODE ) {
		// fill structure with default values
		Net_sendArgsInit( &npsargs );

		// fill network header custom values
		npsargs.netheader.to = GATEWAY_NODE;				// receiver's address
		npsargs.netheader.type = sht11_data_packet_type;	// application protocol
		npsargs.netheader.flags = NET_TLFLAG_ACKREQ;	// network layer config flags (CRC
		npsargs.payload[0].buffer = &data;
		npsargs.payload[0].size = sizeof(struct sht11_data);
	
		// queue packet for sending
		Net_send( &npsargs );
	}
	
	Print_Sht11( &Configuration.id );
	Timers_add(SHT11_INTERVAL, &Send_Sht11, NULL);
}

void Print_Sht11(netaddr_t* source) {
	uint8_t	buf[10];
	ftoa(buf, data.temperature, 0);
	printf("[%u:sht11] %s", *source, buf);
	ftoa(buf, data.humidity, 0);
	printf("C %s%%RH\r\n", buf);
}

void Process_init()
{
	System_registerCallback( C_RADIO, ( fp_vp )Process_radioHandler );
	System_registerCallback( C_SENSOR, ( fp_vp )Process_sensorHandler );
	
	if( Configuration.id != GATEWAY_NODE ) {
		Net_changeMode(NET_TXMODE_LP);
	}
	Timers_add(3000, &Send_Sht11, NULL);
}

/**
 * @brief	Ping request
 *
 * Send a ping request to another node
 * Example: ping 24
 */
COMMAND( ping, 0, cmdargs )
{
	uint16_t node;

	node = String_parseUint16( cmdargs->args, NULL );
	if ( node > 0 ) {
		Net_sendPing((netaddr_t)node);
	}
}

/**
 * @brief Echo Arguments
 *
 * Testcase 1: long command name
 * Testcase 2: argument splitting
 */
COMMAND( echoargs, CMDFLAG_SERIAL, cmdargs )
{
	uint16_t i;
	uint16_t count = Cmd_SplitArgs( cmdargs->args );
	char* arg = (char*) cmdargs->args;

	for ( i = 1; i <= count; i++ ) {
		printf("%u: %*s\n", i, strlen(arg), arg);
		arg = ( char* )Cmd_GetNextArg( arg );
	}
	snprintf(cmdargs->response_args, cmdargs->response_bufsize, "%u args found", count );
}

// Software header.
APP(
    APPH_STATIC_APP,	// Application Type
    __APP_NAME,			// Application Name (8 chars)
    1, 0				// Aplication Version (2 bytes)
);
