#ifndef FAT16_H_
#define FAT16_H_

#define FAT16_SEC				uint16_t	// 16 bit sector addresses
#define FAT32_SEC				uint32_t	// 32 bit sector addresses

typedef enum FAT_TYPE {
	FAT12	= 4,
	FAT16	= 1,
	FAT32	= 2
} FAT_TYPE_T;

#if FAT32_SUPPORT
	#error "FAT32 is currently not supported"
	typedef FAT32_SEC			fat_sa_t;
	typedef FAT32_SEC			fat_ca_t;
	#define FAT_ENTRY_WIDTH(pFS)	(( pFS ->FATType == FAT12) ? FAT16 : pFS ->FATType)
#else
	typedef FAT16_SEC			fat_sa_t;
	typedef FAT16_SEC			fat_ca_t;
	#define FAT_ENTRY_WIDTH(pFS)	(FAT16)
#endif

typedef uint16_t	fat_date_t;
typedef uint16_t	fat_time_t;

#ifdef __TI_COMPILER_VERSION__
	#pragma pack(1)
#endif

struct fat_direntry {
	char		Name[11];
	uint8_t		Attributes;
	uint8_t		NTRes;				///< "set to 0 when created and never modify or look at it after that"
	uint8_t		CrtTimeTenth;
	fat_time_t	CrtTime;
	fat_date_t	CrtDate;
	fat_date_t	LstAccDate;
	uint16_t	FstClusHI;
	fat_time_t	WrtTime;
	fat_date_t	WrtDate;
	uint16_t	FstClusLO;
	uint32_t	FileSize;
}
#if defined(__MSP430GCC__)
	__attribute__((packed))
#endif
;

#ifdef __TI_COMPILER_VERSION__
	#pragma pack()
#endif

/**
 * @brief FAT state structure
 * Information to keep in RAM during use of FAT module. Modified and cut down BPB
 * for any FAT type plus some additions precalculated adresses.
 * 
 * @TODO: check against FAT32 def.
 */
typedef struct fat_state {
	uint8_t		BytsPerSec_bit;		///< Bytes per Sector bitshift value
	uint8_t		SecPerClus_bit;		///< sectors per cluster bitshift value
	uint8_t		BytsPerClus_bit;	///< Bytes per Cluster bitshift value
	uint8_t		NumFATs;			///< Number of FAT (usually 2)
	uint16_t	RsvdSecCnt;			///< Reserved sectors
	uint16_t	RootEntCnt;			///< Number of root directory entries
	uint32_t	TotSec;				///< Total count of sectors
	fat_sa_t	FATSz;				///< Number of sectors occupied by one FAT copy
	
	FAT_TYPE_T	FATType;			///< FAT12 / FAT16 / FAT32
	uint16_t	RootDirSz;			///< size of root dir in sectors
	fat_sa_t	FatSA;				///< first sector of FAT
	uint32_t	FatA;				///< byte address of FAT
	fat_sa_t	FirstClusterSA;		///< first sector of cluster 0
	fat_ca_t	RootDirCA;			///< first cluster of root dir
	
   #if FAT_WRITE
	fat_ca_t	LastFreeCA;			///< first sector of searched block
   #endif
} fat_state_t;


/**
 * @name File handles
 */
typedef enum file_handle_flag {
	FILE_HANDLE_FLAG_NONE		= 0x00u,
	FILE_HANDLE_FLAG_ROOT		= 0x08u,
	FILE_HANDLE_FLAG_DIRECTORY	= 0x10u,	// 	=FAT_ATTR_DIRECTORY
	FILE_HANDLE_FLAG_ROOTDIR	= 0x18u,
	FILE_HANDLE_FLAG_EOF		= 0x80u
} file_handle_flag_t;

typedef struct file_handle {
	fat_ca_t	StartCA;
	fat_ca_t	CurrentCA;
	uint16_t	ClusterOffset;
	uint32_t	Position;
	uint32_t	Size;
	uint32_t	DirEntryPos;
	file_handle_flag_t	Flags;
} file_handle_t;
///@}

typedef struct fat_find_mask {
	uint8_t		AttribMask;
	char		Name[11];
} fat_find_mask_t;

enum fat_init_result {
	FAT_ERROR_NONE,
	FAT_ERROR_SPEC,
	FAT_ERROR_TYPE,
	FAT_ERROR_READ,
	FAT_STRUCT_PACKING
};

#define FAT_DIRENTRY_SIZE		32

#define FAT_NAME_FLAG_DELETED	0xE5u		///< entry free
#define FAT_NAME_FLAG_END		0x00u		///< entry free and no more to come
#define FAT_NAME_FLAG_E5		0x05u		///< this char should be kanjii char 0xE5 but can't be

#define FAT_DIRINDEX_THIS		0
#define FAT_DIRINDEX_VOLID		0
#define FAT_DIRINDEX_PARENT		1

/**
 * @name File attributes as found in DirEntry structure
 */
#define FAT_ATTR_READ_ONLY		0x01u
#define FAT_ATTR_HIDDEN			0x02u
#define FAT_ATTR_SYSTEM			0x04u
#define FAT_ATTR_VOLUME_ID		0x08u
#define FAT_ATTR_DIRECTORY		0x10u
#define FAT_ATTR_ARCHIVE		0x20u
#define FAT_ATTR_LONG_NAME		(FAT_ATTR_READ_ONLY | FAT_ATTR_HIDDEN | FAT_ATTR_SYSTEM | FAT_ATTR_VOLUME_ID)
#define FAT_ATTR_ANY_FILE		(FAT_ATTR_READ_ONLY | FAT_ATTR_HIDDEN | FAT_ATTR_SYSTEM | FAT_ATTR_ARCHIVE)
#define FAT_ATTR_FILE			(FAT_ATTR_READ_ONLY | FAT_ATTR_ARCHIVE)
///@}

/**
 * @name Public Interface
 */
enum fat_init_result Fat_Init(fat_state_t* pFS);

void Fat_GetRootDir(const fat_state_t* pFS, file_handle_t* hRoot);

/// Calculate file handle from FAT directory entry structure
void Fat_DirEntryToHandle(file_handle_t* hDir, uint32_t* pAddress, struct fat_direntry* dirEntry);

uint16_t Fat_ReadFile(
		const fat_state_t*	pFS,
		file_handle_t*		hFile,
		void*				pBuffer,
		uint16_t				count
);
						  
bool Fat_GetNextDirEntry(
		const fat_state_t*	pFS,
		file_handle_t*		hParent,
		uint32_t*				pAddress,
		struct fat_direntry*		resultDirEntry
);

#if FAT_FIND_SIMPLE
	/// Matching function for comparing a 11 char file name against a match string
	bool Fat_MatchName(const char* fatName, const char* match);
	bool Fat_FindFileSimple(
			const fat_state_t*	pFS,
			file_handle_t*		hParent,
			const char*			name,
			const uint8_t			attribMask,
			struct fat_direntry*		resultDirEntry,
			file_handle_t*  	hResultFile		// optional
	);
#endif

//#if FAT_CHANGEDIR
	/// Implementation of the CD command to change current directory by name
	bool Fat_ChangeDir(fat_state_t*	pFS, file_handle_t* hCurDir, char* name);
//#endif

void File_Close( file_handle_t* hFile );
void File_Reset( file_handle_t* hFile );

/// Move file handle a number of bytes forward
void File_Seek(
	const fat_state_t*	pFS,
	file_handle_t*		hFile,
	const uint32_t		count,
	unsigned long		*pAddress
);

#if FAT_FILE_HELPER
	void Fat_Date(fat_date_t* date, uint16_t year, uint8_t month, uint8_t day);
	void Fat_Time(fat_time_t* time, uint8_t hours, uint8_t minutes, uint8_t seconds);
	void Fat_CanonicalNameToFAT(char* fatName, const char* fileName);
	#if FAT_FIND_SIMPLE
		bool Fat_FileExists(
			const fat_state_t* pFS,
			file_handle_t* hParent,
			const char* fileName
		);
	#endif
#endif

#if FAT_WRITE
	bool Fat_UpdateDirEntry(
		fat_state_t* pFS,
		uint32_t* pAddress,
		struct fat_direntry* pDirEntry
	);
	bool Fat_CreateFile(
		fat_state_t*		pFS,
		file_handle_t*		hParent,
		const char*			name,
		const uint8_t			attributes,
		const fat_date_t	date,
		const fat_time_t	time,
		const uint32_t		size,
		file_handle_t*		hFile,
		struct fat_direntry*		pDirEntry
	);
	/**
	 * @brief Erase contents of cluster
	 */
	bool Fat_EraseCluster(fat_state_t* pFS, fat_ca_t cluster);
	/**
	 * @brief Erase contents of chained clusters
	 */
	bool Fat_EraseChain(fat_state_t* pFS, fat_ca_t firstCA);
#endif

#endif /*FAT16_H_*/
