#ifndef FAT_INTERNALS_H_
#define FAT_INTERNALS_H_

#include <string.h>
#include <msp430x16x.h>
#include <ScatterWeb.TypeDefs.h>
#include <ScatterWeb.Sd.h>
#include <ScatterWeb.Fat.h>

#define RETF(x)		if( ! x ) return false;
#define	MIN(x, y)	( x < y ) ? x : y 

#ifdef __TI_COMPILER_VERSION__
	#pragma pack(1)
#endif

/**
 * @brief MBR Partition entry
 */
struct fat_partition {
	uint8_t	state;
	uint8_t	firsthead;
	uint16_t	firstsector		:  6;
	uint16_t	firstcylinder	: 10;
	uint8_t	type;
	uint8_t	lasthead;
	uint16_t	lastsector		:  6;
	uint16_t	lastcylinder	: 10;
	uint32_t	distance_to_first;
	uint32_t	numsectors;
} // 16 bytes
#if defined(__MSP430GCC__)
	__attribute__((packed))
#endif
;

/**
 * @brief FAT BIOS Parameter Block
 * Common for all FAT types
 */
struct fat_bpb {
	uint16_t	BytsPerSec;
	uint8_t		SecPerClus;
	uint16_t	RsvdSecCnt;
	uint8_t		NumFATs;
	uint16_t	RootEntCnt;
	uint16_t	TotSec16;
	uint8_t		Media;
	uint16_t	FATSz16;
	uint16_t	SecPerTrk;
	uint16_t	NumHeads;
	uint32_t	HiddSec;
	uint32_t	TotSec32;	
} // 25 bytes
#if defined(__MSP430GCC__)
	__attribute__((packed))
#endif
;

/**
 * @brief FAT BIOS Parameter Block Extension for FAT12 and FAT16
 */
struct fat16_bpbext {
	// ignore boot sector
	//  char		jump_code[3];
	//  char		oem_name[8];
	// BIOS parameter block
	// fat16 specific
	uint16_t	logical_drive_number;
	uint8_t	ext_signature;
	uint32_t	serial_number;
	char	volume_name[11];
	char	fat_name[8];
}
#if defined(__MSP430GCC__)
	__attribute__((packed))
#endif
;

/**
 * @brief FAT BIOS Parameter Block Extension for FAT32
 * @todo: for fat32 support
 */
struct fat32_bpbext {
	// ignore boot sector
	//  char		jump_code[3];
	//  char		oem_name[8];
	// BIOS parameter block
	// fat32 specific
	char	raw[46];
	char	fat_name[8];
}
#if defined(__MSP430GCC__)
	__attribute__((packed))
#endif
;

/**
 * @brief FAT Structure
 * FAT12, FAT16, FAT32
 */
struct fat {
	struct fat_bpb				bpb;
	union {
		struct fat16_bpbext		fat12;
		struct fat16_bpbext		fat16;
#if FAT32_SUPPORT
		struct fat32_bpbext		fat32;
#endif
	} vbr;
}
#if defined(__MSP430GCC__)
	__attribute__((packed))
#endif
;

#ifdef __TI_COMPILER_VERSION__
	#pragma pack()
#endif

#define FAT_BIOS_SIGNATURE		0xAA55		///< A multi-purpose magical number
#define FAT_EOC12				0x0FF8		///< FAT12 last cluster symbol
#define FAT_EOC16				0xFFF8		///< FAT16 last cluster symbol
#define FAT_EOC32				0x0FFFFFF8	///< FAT32 last cluster symbol
#define FAT_BAD12				0x0FF7
#define FAT_BAD16				0xFFF7
#define FAT_BAD32				0x0FFFFFF7
#define FAT_FREE				0
#define	FAT_BPB_OFFSET			11			///< offset from VBR start to bpb in bytes
#define	FAT_CLUSTER_OFFSET		2			///< reserved entries at beginning of FAT

/**
 * @name Private Interface
 */
/// Returns 0-based number of lowest bit set
uint8_t get_lowest_bit_index(int i);

/// Convert file handle to absolute address
uint32_t _Fat_GetFilePos(const fat_state_t* pFS, const file_handle_t* hFile);

/// Calculate FAT type (the specified way)
FAT_TYPE_T _Fat_GetType(const fat_state_t* pFS, struct fat* pFAT);

/// Calculate byte address of fat entry from cluster address
void _Fat_getClusterFAT(const fat_state_t* pFS, fat_ca_t* pCA, uint32_t* pAddress);
/// Calculate byte address from cluster address 
void _Fat_getClusterPos(const fat_state_t* pFS, const fat_ca_t* pCA, uint32_t* pAddress);
bool _fat_getNextCA(const fat_state_t* pFS, fat_ca_t cluster, fat_ca_t* pNextCluster);

/// Test a file handle for validity
#define FILE_HANDLE_INVALID(x)	( x ->StartCA == 0)

#if FAT_WRITE
	bool _Fat_GetChain(fat_state_t* pFS, fat_ca_t* pFirstCA, uint16_t count);
	bool _Fat_Update(fat_state_t* pFS, fat_ca_t cluster, fat_ca_t nextCluster);
#endif

///@}

#endif /*FAT_INTERNALS_H_*/
